@@ -426,7 +426,12 @@ class Sparse3DNA(nn.Module):
         self.padding_height = calc_same_padding(self.kernel_size[1], self.dilation[1])
         self.padding_width = calc_same_padding(self.kernel_size[2], self.dilation[2])
 
-        self.video_padding = (self.padding_width, self.padding_width, self.padding_height, self.padding_height, self.padding_frame, self.padding_frame)
+        # use separate padding for causal convolution vs non-causal
+
+        if causal:
+            self.video_padding = (self.padding_width * 2, 0, self.padding_height * 2, 0, self.padding_frame * 2, 0)
+        else:
+            self.video_padding = (self.padding_width, self.padding_width, self.padding_height, self.padding_height, self.padding_frame, self.padding_frame)
 
         # save video shape and calculate max number of tokens
 
@@ -441,21 +446,19 @@ class Sparse3DNA(nn.Module):
 
         # precalculate causal mask
 
-        indices = torch.arange(max_num_tokens)
-        shaped_indices = rearrange(indices, '(f h w) -> 1 1 f h w', f = max_frames, h = fmap_size, w = fmap_size)
-        padded_indices = F.pad(shaped_indices, self.video_padding, value = max_num_tokens) # padding has value of max tokens so to be masked out
-        unfolded_indices = unfoldNd(padded_indices, kernel_size = self.kernel_size, dilation = self.dilation)
-        unfolded_indices = rearrange(unfolded_indices, '1 k n -> n k')
+        ones = torch.ones((max_num_tokens,))
+        ones = rearrange(ones, '(f h w) -> 1 1 f h w', f = max_frames, h = fmap_size, w = fmap_size)
+        ones = F.pad(ones, self.video_padding, value = 0.)
+        ones = unfoldNd(ones, kernel_size = self.kernel_size, dilation = self.dilation)
+        ones = rearrange(ones, '1 k n -> n k')
 
-        # if causal, compare query and key indices and make sure past cannot see future
-        # if not causal, just mask out the padding
+        # mask out padding
 
-        if causal:
-            mask = rearrange(indices, 'n -> n 1') < unfolded_indices
-        else:
-            mask = unfolded_indices == max_num_tokens
+        padding_mask = ones == 0.
 
-        mask = F.pad(mask, (1, 0), value = False) # bos tokens never get masked out
+        # bos tokens never get masked out
+
+        mask = F.pad(padding_mask, (1, 0), value = False)
         self.register_buffer('mask', mask)
 
     def forward(self, x, **kwargs):
@@ -643,34 +646,32 @@ class SparseCausal2DNA(nn.Module):
 
         self.kernel_size = (kernel_size, height)
         self.dilation = (dilation, 1)
-        self.padding = (calc_same_padding(kernel_size, dilation), 0)
 
+        self.causal_padding = (0, 0, calc_same_padding(kernel_size, dilation) * 2, 0)
         self.rel_pos_bias = AxialPositionalEmbedding(heads, shape = self.kernel_size) if exists(rel_pos_bias) else None
 
         # causal mask
 
-        self.register_buffer('causal_mask', None, persistent = False)
+        self.register_buffer('mask', None, persistent = False)
 
-    def get_causal_mask(self, t):
-        if exists(self.causal_mask) and self.causal_mask.shape[-3] == t.shape[-3]:
-            return self.causal_mask
+    def get_mask(self, t):
+        if exists(self.mask) and self.mask.shape[-3] == t.shape[-3]:
+            return self.mask
 
         device, seq_len = t.device, t.shape[-3] * self.height
-        q_range = torch.arange(seq_len, device = device, dtype = torch.float32)
-        k_range = torch.arange(seq_len, device = device, dtype = torch.float32)
 
-        q_range = rearrange(q_range, '(n m) -> n m', m = self.height)
-        k_range = rearrange(k_range, '(n m) -> 1 1 n m', m = self.height)
+        ones = torch.ones((seq_len,), device = device)
+        ones = rearrange(ones, '(n m) -> 1 1 n m', m = self.height)
 
-        k_range = F.pad(k_range, (0, 0, self.padding[0], self.padding[0]), value = seq_len)
-        k_range = unfoldNd(k_range, kernel_size = self.kernel_size, dilation = self.dilation)
-        k_range = rearrange(k_range, '1 d n -> n d')
+        ones = F.pad(ones, self.causal_padding, value = 0.)
+        ones = unfoldNd(ones, kernel_size = self.kernel_size, dilation = self.dilation)
+        ones = rearrange(ones, '1 d n -> n d')
 
-        causal_mask = rearrange(q_range, 'n i -> n i 1') < rearrange(k_range, 'n j -> n 1 j')
-        causal_mask = F.pad(causal_mask, (1, 0), value = False)
+        padding_mask = rearrange(ones, 'n j -> n 1 j') == 0.
+        mask = F.pad(padding_mask, (1, 0), value = False)
 
-        self.register_buffer('causal_mask', causal_mask, persistent = False)
-        return causal_mask
+        self.register_buffer('mask', mask, persistent = False)
+        return mask
 
     def forward(
         self,
@@ -717,7 +718,8 @@ class SparseCausal2DNA(nn.Module):
         # reshape key / values to be unfolded
 
         k, v = map(lambda t: rearrange(t, 'b h (x y) d -> (b h) d x y ', y = tokens_per_timestep), (k, v))
-        k, v = map(lambda t: F.unfold(t, kernel_size = self.kernel_size, dilation = self.dilation, padding = self.padding), (k, v))
+        k, v = map(lambda t: F.pad(t, self.causal_padding), (k, v))
+        k, v = map(lambda t: F.unfold(t, kernel_size = self.kernel_size, dilation = self.dilation), (k, v))
         k, v = map(lambda t: rearrange(t, '(b h f) (d j) i -> b h i (f j) d', b = b, h = h, j = kernel_numel), (k, v))
 
         # add bos
@@ -741,8 +743,8 @@ class SparseCausal2DNA(nn.Module):
         # causal + padding mask
 
         mask_value = -torch.finfo(x.dtype).max
-        causal_mask = self.get_causal_mask(sim)
-        sim = sim.masked_fill(causal_mask, mask_value)
+        mask = self.get_mask(sim)
+        sim = sim.masked_fill(mask, mask_value)
 
         # attention
 
@@ -1317,6 +1319,7 @@ class DualModalityDecoder(nn.Module):
         sparse_3dna_kernel_size = 3,
         sparse_3dna_query_num_frames_chunk = None,
         sparse_3dna_dilations = (1,),
+        sparse_3dna_rel_pos_bias = False,
         sparse_2dna_kernel_size = 7,
         sparse_2dna_dilation = (1,),
         sparse_2dna_rel_pos_bias = False,
@@ -1340,7 +1343,8 @@ class DualModalityDecoder(nn.Module):
                 kernel_size = sparse_3dna_kernel_size,
                 dilation = dilation,
                 video_shape = sparse_3dna_video_shape,
-                query_num_frames_chunk = sparse_3dna_query_num_frames_chunk
+                query_num_frames_chunk = sparse_3dna_query_num_frames_chunk,
+                rel_pos_bias = sparse_3dna_rel_pos_bias,
             )
 
             cross_attn = Attention(
@@ -1505,6 +1509,7 @@ class ReversibleDualModalityDecoder(nn.Module):
         sparse_3dna_kernel_size = 3,
         sparse_3dna_query_num_frames_chunk = None,
         sparse_3dna_dilations = (1,),
+        sparse_3dna_rel_pos_bias = False,
         sparse_2dna_kernel_size = 7,
         sparse_2dna_dilation = (1,),
         sparse_2dna_rel_pos_bias = False,
@@ -1532,7 +1537,8 @@ class ReversibleDualModalityDecoder(nn.Module):
                 kernel_size = sparse_3dna_kernel_size,
                 dilation = video_dilation,
                 video_shape = sparse_3dna_video_shape,
-                query_num_frames_chunk = sparse_3dna_query_num_frames_chunk
+                query_num_frames_chunk = sparse_3dna_query_num_frames_chunk,
+                rel_pos_bias = sparse_3dna_rel_pos_bias,
             )
 
             audio_self_attn = SparseCausal2DNA(
@@ -1995,6 +2001,7 @@ class NUWAVideoAudio(nn.Module):
         sparse_3dna_kernel_size = 3,
         sparse_3dna_query_num_frames_chunk = None,
         sparse_3dna_dilation = 1,
+        sparse_3dna_rel_pos_bias = True,
         sparse_2dna_kernel_size = 7,
         sparse_2dna_dilation = 1,
         sparse_2dna_rel_pos_bias = True,
@@ -2074,11 +2081,12 @@ class NUWAVideoAudio(nn.Module):
             sparse_3dna_kernel_size = sparse_3dna_kernel_size,
             sparse_3dna_dilations = sparse_3dna_dilations,
             sparse_3dna_query_num_frames_chunk = sparse_3dna_query_num_frames_chunk,
-            sparse_2dna_kernel_size = sparse_2dna_kernel_size,
-            sparse_2dna_dilation = sparse_2dna_dilation,
+            sparse_3dna_rel_pos_bias = sparse_3dna_rel_pos_bias,
             num_audio_tokens_per_video_frame = num_audio_tokens_per_video_frame,
             num_video_tokens_per_frame = fmap_size * fmap_size,
             cross_modality_attn_every = cross_modality_attn_every,
+            sparse_2dna_kernel_size = sparse_2dna_kernel_size,
+            sparse_2dna_dilation = sparse_2dna_dilation,
             sparse_2dna_rel_pos_bias = sparse_2dna_rel_pos_bias
         )
 
